@@ -13,31 +13,46 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-import autopep8
+from datetime import datetime
+import json
+import logging
 import os
 import re
 import tempfile
 import time
-import requests
-import json
+from typing import Dict
 
-from black import format_str, FileMode
-from datetime import datetime
-from elyra._version import __version__
-from elyra.metadata import MetadataManager
-from elyra.pipeline import RuntimePipelineProcess, PipelineProcessor, PipelineProcessorResponse
-from elyra.util.path import get_absolute_path
-from jinja2 import Environment, PackageLoader
+import autopep8
+from black import FileMode
+from black import format_str
+from jinja2 import Environment
+from jinja2 import PackageLoader
+from jupyter_core.paths import ENV_JUPYTER_PATH
 from kfp import Client as ArgoClient
 from kfp import compiler as kfp_argo_compiler
-from kfp.aws import use_aws_secret
-from kfp_tekton import TektonClient, compiler as kfp_tekton_compiler
-from kfp_notebook.pipeline import NotebookOp
+from kfp import components as components
+from kfp.aws import use_aws_secret  # noqa H306
 from kfp_server_api.exceptions import ApiException
-from urllib3.exceptions import LocationValueError, MaxRetryError
+from kfp_tekton import compiler as kfp_tekton_compiler
+from kfp_tekton import TektonClient
+import requests
+from urllib3.exceptions import LocationValueError
+from urllib3.exceptions import MaxRetryError
+
+from elyra._version import __version__
+from elyra.kfp.operator import ExecuteFileOp
+from elyra.metadata.manager import MetadataManager
+from elyra.pipeline.component_parser_kfp import KfpComponentParser
+from elyra.pipeline.pipeline import GenericOperation
+from elyra.pipeline.pipeline import Operation
+from elyra.pipeline.processor import PipelineProcessor
+from elyra.pipeline.processor import PipelineProcessorResponse
+from elyra.pipeline.processor import RuntimePipelineProcessor
+from elyra.util.path import get_absolute_path
+from elyra.util.path import get_expanded_path
 
 
-class KfpPipelineProcessor(RuntimePipelineProcess):
+class KfpPipelineProcessor(RuntimePipelineProcessor):
     _type = 'kfp'
 
     # Provide users with the ability to identify a writable directory in the
@@ -50,6 +65,9 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
     def type(self):
         return self._type
 
+    def __init__(self, root_dir, **kwargs):
+        super().__init__(root_dir, component_parser=KfpComponentParser(), **kwargs)
+
     def process(self, pipeline):
         """Runs a pipeline on Kubeflow Pipelines
 
@@ -63,7 +81,7 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
         runtime_configuration = self._get_metadata_configuration(namespace=MetadataManager.NAMESPACE_RUNTIMES,
                                                                  name=pipeline.runtime_config)
 
-        api_endpoint = runtime_configuration.metadata['api_endpoint']
+        api_endpoint = runtime_configuration.metadata['api_endpoint'].rstrip('/')
         cos_endpoint = runtime_configuration.metadata['cos_endpoint']
         cos_bucket = runtime_configuration.metadata['cos_bucket']
 
@@ -77,23 +95,36 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
 
         pipeline_name = pipeline.name
         try:
-            # Connect to the KFP server and determine whether
-            # a pipeline with this name was already uploaded
-            session_cookie = None
+            # Connect to the Kubeflow server, determine whether it is secured,
+            # and if it is try to authenticate with the user-provided credentials
+            # (if any were defined in the runtime configuration)
 
-            if api_username and api_password:
-                endpoint = api_endpoint.replace('/pipeline', '')
-                session_cookie = self._get_user_auth_session_cookie(endpoint,
-                                                                    api_username,
-                                                                    api_password)
+            endpoint = api_endpoint.replace('/pipeline', '')
+            auth_info = \
+                KfpPipelineProcessor._get_user_auth_session_cookie(endpoint,
+                                                                   api_username,
+                                                                   api_password)
+
+            self.log.debug(f"Kubeflow authentication info: {auth_info}")
+
+            if auth_info['endpoint_secured'] and \
+               auth_info['authservice_session_cookie'] is None:
+                # Kubeflow is secured but our attempt to authenticate did
+                # not yield the expected results. Log the collected authentication
+                # information and abort processing.
+                self.log.warning(f"Kubeflow authentication info: {auth_info}")
+                raise RuntimeError(f"Error connecting to Kubeflow at '{endpoint}'"
+                                   f": Authentication request failed. Check the "
+                                   f"Kubeflow Pipelines credentials in runtime "
+                                   f"configuration '{pipeline.runtime_config}'.")
 
             # Create a KFP client
             if 'Tekton' == engine:
                 client = TektonClient(host=api_endpoint,
-                                      cookies=session_cookie)
+                                      cookies=auth_info['authservice_session_cookie'])
             else:
                 client = ArgoClient(host=api_endpoint,
-                                    cookies=session_cookie)
+                                    cookies=auth_info['authservice_session_cookie'])
 
             # Determine whether a pipeline with the provided
             # name already exists
@@ -114,11 +145,15 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
             # will be stored
             cos_directory = f'{pipeline_name}-{timestamp}'
 
-        except MaxRetryError as ex:
-            raise RuntimeError('Error connecting to pipeline server {}'.format(api_endpoint)) from ex
+        except (requests.exceptions.ConnectionError, MaxRetryError) as ce:
+            raise RuntimeError(f"Error connecting to pipeline server {api_endpoint}.  Check the "
+                               f"Kubeflow Pipelines connection information in runtime "
+                               f"configuration '{pipeline.runtime_config}'.") from ce
         except LocationValueError as lve:
             if api_username:
-                raise ValueError("Failure occurred uploading pipeline, check your credentials") from lve
+                raise ValueError(f"Failure occurred uploading pipeline. Check the "
+                                 f"Kubeflow Pipelines credentials in runtime "
+                                 f"configuration '{pipeline.runtime_config}'.") from lve
             else:
                 raise lve
 
@@ -158,7 +193,9 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
                 else:
                     kfp_argo_compiler.Compiler().compile(pipeline_function, pipeline_path)
             except Exception as ex:
-                raise RuntimeError('Error compiling pipeline {} for engine {} at {}'.
+                if ex.__cause__:
+                    raise RuntimeError(str(ex)) from ex
+                raise RuntimeError('Error pre-processing pipeline {} for engine {} at {}'.
                                    format(pipeline_name, engine, pipeline_path), str(ex)) from ex
 
             self.log.debug("Kubeflow Pipeline was created in %s", pipeline_path)
@@ -245,10 +282,12 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
 
         runtime_configuration = self._get_metadata_configuration(namespace=MetadataManager.NAMESPACE_RUNTIMES,
                                                                  name=pipeline.runtime_config)
-        api_endpoint = runtime_configuration.metadata['api_endpoint']
+        api_endpoint = runtime_configuration.metadata['api_endpoint'].rstrip('/')
         namespace = runtime_configuration.metadata.get('user_namespace')
         engine = runtime_configuration.metadata.get('engine')
         cos_secret = runtime_configuration.metadata.get('cos_secret')
+        kf_secured = runtime_configuration.metadata.get('api_username') is not None and \
+            runtime_configuration.metadata.get('api_password') is not None
 
         if os.path.exists(absolute_pipeline_export_path) and not overwrite:
             raise ValueError("File " + absolute_pipeline_export_path + " already exists.")
@@ -260,9 +299,9 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
                 # Exported pipeline is not associated with an experiment
                 # or a version. The association is established when the
                 # pipeline is imported into KFP by the user.
-                pipeline_function = lambda: self._cc_pipeline(pipeline,  # nopep8
-                                                              '')
-
+                pipeline_function = lambda: self._cc_pipeline(pipeline,
+                                                              pipeline_name,
+                                                              cos_directory=cos_directory)  # nopep8
                 if 'Tekton' == engine:
                     self.log.info("Compiling pipeline for Tekton engine")
                     kfp_tekton_compiler.TektonCompiler().compile(pipeline_function, absolute_pipeline_export_path)
@@ -270,7 +309,9 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
                     self.log.info("Compiling pipeline for Argo engine")
                     kfp_argo_compiler.Compiler().compile(pipeline_function, absolute_pipeline_export_path)
             except Exception as ex:
-                raise RuntimeError('Error compiling pipeline {} for export at {}'.
+                if ex.__cause__:
+                    raise RuntimeError(str(ex)) from ex
+                raise RuntimeError('Error pre-processing pipeline {} for export at {}'.
                                    format(pipeline_name, absolute_pipeline_export_path), str(ex)) from ex
         else:
             # Export pipeline as Python DSL
@@ -292,14 +333,13 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
 
             description = f'Created with Elyra {__version__} pipeline editor using {pipeline.source}.'
 
-            for key, operation in defined_pipeline.items():
-                self.log.debug("component :\n "
-                               "container op name : %s \n "
-                               "inputs : %s \n "
-                               "outputs : %s \n ",
-                               operation.name,
-                               operation.inputs,
-                               operation.outputs)
+            if self.log.isEnabledFor(logging.DEBUG):
+                self.log.debug(f"Exporting pipeline {pipeline_name} with components: \n")
+                for key, operation in defined_pipeline.items():
+                    self.log.debug("component:\n "
+                                   f"operation name : {operation.name} \n "
+                                   f"inputs : {operation.inputs} \n "
+                                   f"outputs : {operation.outputs} \n ")
 
             # The exported pipeline is by default associated with
             # an experiment.
@@ -316,7 +356,8 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
                                             namespace=namespace,
                                             api_endpoint=api_endpoint,
                                             pipeline_description=description,
-                                            writable_container_dir=self.WCD)
+                                            writable_container_dir=self.WCD,
+                                            kf_secured=kf_secured)
 
             # Write to Python file and fix formatting
             with open(absolute_pipeline_export_path, "w") as fh:
@@ -332,6 +373,17 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
 
         return pipeline_export_path  # Return the input value, not its absolute form
 
+    def _collect_envs(self, operation: Operation, **kwargs) -> Dict:
+        """
+        Amends envs collected from superclass with those pertaining to this subclass
+
+        :return: dictionary containing environment name/value pairs
+        """
+        envs = super()._collect_envs(operation, **kwargs)
+        # Only Unix-style path spec is supported.
+        envs['ELYRA_WRITABLE_CONTAINER_DIR'] = self.WCD
+        return envs
+
     def _cc_pipeline(self,
                      pipeline,
                      pipeline_name,
@@ -347,6 +399,7 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
         cos_username = runtime_configuration.metadata['cos_username']
         cos_password = runtime_configuration.metadata['cos_password']
         cos_secret = runtime_configuration.metadata.get('cos_secret')
+        engine = runtime_configuration.metadata['engine']
 
         if cos_directory is None:
             cos_directory = pipeline_name
@@ -361,7 +414,7 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
         container_runtime = bool(os.getenv('CRIO_RUNTIME', 'False').lower() == 'true')
 
         # Create dictionary that maps component Id to its ContainerOp instance
-        notebook_ops = {}
+        target_ops = {}
 
         # Sort operations based on dependency graph (topological order)
         sorted_operations = PipelineProcessor._sort_operations(pipeline.operations)
@@ -375,92 +428,143 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
 
         for operation in sorted_operations:
 
-            operation_artifact_archive = self._get_dependency_archive_name(operation)
-
-            self.log.debug("Creating pipeline component :\n {op} archive : {archive}".format(
-                           op=operation, archive=operation_artifact_archive))
-
             if container_runtime:
                 # Volume size to create when using CRI-o, NOTE: IBM Cloud minimum is 20Gi
                 emptydir_volume_size = '20Gi'
 
-            # Collect env variables
-            pipeline_envs = dict()
-            if not cos_secret:
-                pipeline_envs['AWS_ACCESS_KEY_ID'] = cos_username
-                pipeline_envs['AWS_SECRET_ACCESS_KEY'] = cos_password
-            # Convey pipeline logging enablement to operation
-            pipeline_envs['ELYRA_ENABLE_PIPELINE_INFO'] = str(self.enable_pipeline_info)
-            # Setting identifies a writable directory in the container image.
-            # Only Unix-style path spec is supported.
-            pipeline_envs['ELYRA_WRITABLE_CONTAINER_DIR'] = self.WCD
-
-            if operation.env_vars:
-                for env_var in operation.env_vars:
-                    # Strip any of these special characters from both key and value
-                    # Splits on the first occurrence of '='
-                    result = [x.strip(' \'\"') for x in env_var.split('=', 1)]
-                    # Should be non empty key with a value
-                    if len(result) == 2 and result[0] != '':
-                        pipeline_envs[result[0]] = result[1]
-
             sanitized_operation_name = self._sanitize_operation_name(operation.name)
 
-            # create pipeline operation
-            notebook_ops[operation.id] = NotebookOp(name=sanitized_operation_name,
-                                                    pipeline_name=pipeline_name,
-                                                    experiment_name=experiment_name,
-                                                    notebook=operation.filename,
-                                                    cos_endpoint=cos_endpoint,
-                                                    cos_bucket=cos_bucket,
-                                                    cos_directory=cos_directory,
-                                                    cos_dependencies_archive=operation_artifact_archive,
-                                                    pipeline_version=pipeline_version,
-                                                    pipeline_source=pipeline.source,
-                                                    pipeline_inputs=operation.inputs,
-                                                    pipeline_outputs=operation.outputs,
-                                                    pipeline_envs=pipeline_envs,
-                                                    emptydir_volume_size=emptydir_volume_size,
-                                                    cpu_request=operation.cpu,
-                                                    mem_request=operation.memory,
-                                                    gpu_limit=operation.gpu,
-                                                    image=operation.runtime_image,
-                                                    file_outputs={
-                                                        'mlpipeline-metrics':
-                                                            '{}/mlpipeline-metrics.json'
-                                                            .format(pipeline_envs['ELYRA_WRITABLE_CONTAINER_DIR']),
-                                                        'mlpipeline-ui-metadata':
-                                                            '{}/mlpipeline-ui-metadata.json'
-                                                            .format(pipeline_envs['ELYRA_WRITABLE_CONTAINER_DIR'])
-                                                    })
+            # Create pipeline operation
+            # If operation is one of the "generic" set of NBs or scripts, construct custom ExecuteFileOp
+            if isinstance(operation, GenericOperation):
 
-            if cos_secret and not export:
-                notebook_ops[operation.id].apply(use_aws_secret(cos_secret))
+                # Collect env variables
+                pipeline_envs = self._collect_envs(operation,
+                                                   cos_secret=cos_secret,
+                                                   cos_username=cos_username,
+                                                   cos_password=cos_password)
 
-            image_namespace = self._get_metadata_configuration(namespace=MetadataManager.NAMESPACE_RUNTIME_IMAGES)
-            for image_instance in image_namespace:
-                if image_instance.metadata['image_name'] == operation.runtime_image and \
-                   image_instance.metadata.get('pull_policy'):
-                    notebook_ops[operation.id].container.set_image_pull_policy(image_instance.metadata['pull_policy'])
+                operation_artifact_archive = self._get_dependency_archive_name(operation)
 
-            self.log_pipeline_info(pipeline_name,
-                                   f"processing operation dependencies for id: {operation.id}",
-                                   operation_name=operation.name)
+                self.log.debug("Creating pipeline component:\n {op} archive : {archive}".format(
+                               op=operation, archive=operation_artifact_archive))
 
-            self._upload_dependencies_to_object_store(runtime_configuration,
-                                                      cos_directory,
-                                                      operation)
+                target_ops[operation.id] = ExecuteFileOp(name=sanitized_operation_name,
+                                                         pipeline_name=pipeline_name,
+                                                         experiment_name=experiment_name,
+                                                         notebook=operation.filename,
+                                                         cos_endpoint=cos_endpoint,
+                                                         cos_bucket=cos_bucket,
+                                                         cos_directory=cos_directory,
+                                                         cos_dependencies_archive=operation_artifact_archive,
+                                                         pipeline_version=pipeline_version,
+                                                         pipeline_source=pipeline.source,
+                                                         pipeline_inputs=operation.inputs,
+                                                         pipeline_outputs=operation.outputs,
+                                                         pipeline_envs=pipeline_envs,
+                                                         emptydir_volume_size=emptydir_volume_size,
+                                                         cpu_request=operation.cpu,
+                                                         mem_request=operation.memory,
+                                                         gpu_limit=operation.gpu,
+                                                         workflow_engine=engine,
+                                                         image=operation.runtime_image,
+                                                         file_outputs={
+                                                             'mlpipeline-metrics':
+                                                                 '{}/mlpipeline-metrics.json'
+                                                                 .format(pipeline_envs['ELYRA_WRITABLE_CONTAINER_DIR']),  # noqa
+                                                             'mlpipeline-ui-metadata':
+                                                                 '{}/mlpipeline-ui-metadata.json'
+                                                                 .format(pipeline_envs['ELYRA_WRITABLE_CONTAINER_DIR'])  # noqa
+                                                         })
+
+                # TODO Can we move all of this to apply to non-standard components as well? Test when servers are up
+                if cos_secret and not export:
+                    target_ops[operation.id].apply(use_aws_secret(cos_secret))
+
+                image_namespace = self._get_metadata_configuration(namespace=MetadataManager.NAMESPACE_RUNTIME_IMAGES)
+                for image_instance in image_namespace:
+                    if image_instance.metadata['image_name'] == operation.runtime_image and \
+                            image_instance.metadata.get('pull_policy'):
+                        target_ops[operation.id].container. \
+                            set_image_pull_policy(image_instance.metadata['pull_policy'])
+
+                self.log_pipeline_info(pipeline_name,
+                                       f"processing operation dependencies for id: {operation.id}",
+                                       operation_name=operation.name)
+
+                self._upload_dependencies_to_object_store(runtime_configuration,
+                                                          cos_directory,
+                                                          operation)
+
+            # If operation is a "non-standard" component, load it's spec and create operation with factory function
+            else:
+                # Retrieve component from cache
+                component = self._component_registry.get_component(operation.classifier)
+
+                # Convert the user-entered value of certain properties according to their type
+                for component_property in component.properties:
+                    # Get corresponding property's value from parsed pipeline
+                    property_value = operation.component_params.get(component_property.ref)
+
+                    self.log.debug(f"Processing component parameter '{component_property.name}' "
+                                   f"of type '{component_property.type}'")
+                    if component_property.type == "file":
+                        filename = get_absolute_path(get_expanded_path(self.root_dir), property_value)
+                        try:
+                            with open(filename) as f:
+                                operation.component_params[component_property.ref] = f.read()
+                        except Exception:
+                            # If file can't be found locally, assume a remote file location was entered.
+                            # This may cause the pipeline run to fail; the user must debug in this case.
+                            pass
+                    elif component_property.type == 'dictionary':
+                        processed_value = self._process_dictionary_value(property_value)
+                        operation.component_params[component_property.ref] = processed_value
+                    elif component_property.type == 'list':
+                        processed_value = self._process_list_value(property_value)
+                        operation.component_params[component_property.ref] = processed_value
+
+                # Get absolute path of component source
+                component_path = component.source
+                if component.source_type == "filename":
+                    component_path = os.path.join(ENV_JUPYTER_PATH[0], 'components', component_path)
+
+                component_source = {component.source_type: component_path}
+
+                # Build component task factory
+                try:
+                    factory_function = components.load_component(**component_source)
+                except Exception as e:
+                    # TODO Fix error messaging and break exceptions down into categories
+                    self.log.error(f"Error loading component spec for {operation.name}: {str(e)}")
+                    raise RuntimeError(f"Error loading component spec for {operation.name}.")
+
+                # Add factory function, which returns a ContainerOp task instance, to pipeline operation dict
+                try:
+                    # Remove inputs and outputs from params dict until support for data exchange is provided
+                    operation.component_params_as_dict.pop("inputs")
+                    operation.component_params_as_dict.pop("outputs")
+
+                    # Create ContainerOp instance and assign appropriate user-provided name
+                    container_op = factory_function(**operation.component_params_as_dict)
+                    container_op.set_display_name(operation.name)
+
+                    target_ops[operation.id] = container_op
+                except Exception as e:
+                    # TODO Fix error messaging and break exceptions down into categories
+                    self.log.error(f"Error constructing component {operation.name}: {str(e)}")
+                    raise RuntimeError(f"Error constructing component {operation.name}.")
 
         # Process dependencies after all the operations have been created
         for operation in pipeline.operations.values():
-            op = notebook_ops[operation.id]
-            for parent_operation_id in operation.parent_operations:
-                parent_op = notebook_ops[parent_operation_id]  # Parent Operation
+            op = target_ops[operation.id]
+            for parent_operation_id in operation.parent_operation_ids:
+                parent_op = target_ops[parent_operation_id]  # Parent Operation
                 op.after(parent_op)
 
         self.log_pipeline_info(pipeline_name, "pipeline dependencies processed", duration=(time.time() - t0_all))
 
-        return notebook_ops
+        return target_ops
 
     @staticmethod
     def _sanitize_operation_name(name: str) -> str:
@@ -471,21 +575,56 @@ class KfpPipelineProcessor(RuntimePipelineProcess):
         return re.sub('-+', '-', re.sub('[^-_0-9A-Za-z ]+', '-', name)).lstrip('-').rstrip('-')
 
     @staticmethod
-    def _get_user_auth_session_cookie(url, username, password):
+    def _get_user_auth_session_cookie(url, username=None, password=None) -> dict:
+        """
+        Determine whether the specified URL is secured by Dex and, if username
+        and password were provided, try to obtain a session cookie. Other forms
+        of authentication are not supported.
+
+        :param url: Kubeflow server URL, including protocol
+        :type url: str
+        :param username: Kubeflow user name, defaults to None
+        :type username: str, optional
+        :param password: Kubeflow user's password, defaults to None
+        :type password: str, optional
+        :return: authentication information
+        :rtype: dict
+        """
+
+        # Return data structure
+        auth_info = {
+            'endpoint': url,                    # KF endpoint URL
+            'endpoint_response_url': None,      # KF redirect URL, if applicable
+            'endpoint_secured': False,          # True if KF is secured [by dex]
+            'authservice_session_cookie': None  # Set if KF is secured & user authenticated
+        }
+
+        # Obtain redirect URL
         get_response = requests.get(url)
 
-        # auth request to kfp server with istio dex look like '/dex/auth/local?req=REQ_VALUE'
-        if 'auth' in get_response.url:
-            credentials = {'login': username, 'password': password}
+        auth_info['endpoint_response_url'] = get_response.url
 
-            # Authenticate user
+        # If KF redirected to '/dex/auth/local?req=REQ_VALUE'
+        # try to authenticate using the provided credentials
+        if 'dex/auth' in get_response.url:
+            auth_info['endpoint_secured'] = True  # KF is secured
+
+            # Try to authenticate user by sending a request to the
+            # Dex redirect URL
             session = requests.Session()
-            session.post(get_response.url, data=credentials)
+            session.post(get_response.url,
+                         data={'login': username,
+                               'password': password})
+            # Capture authservice_session cookie, if one was returned
+            # in the response
             cookie_auth_key = 'authservice_session'
             cookie_auth_value = session.cookies.get(cookie_auth_key)
 
             if cookie_auth_value:
-                return cookie_auth_key + '=' + cookie_auth_value
+                auth_info['authservice_session_cookie'] = \
+                    f"{cookie_auth_key}={cookie_auth_value}"
+
+        return auth_info
 
 
 class KfpPipelineProcessorResponse(PipelineProcessorResponse):
